from collections import defaultdict

from django.db.models import Exists, OuterRef
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied, ValidationError

from arkindex.documents.models import Corpus, Element
from arkindex.ponos.models import Task
from arkindex.process.models import (
    Process,
    ProcessDataset,
    ProcessMode,
    WorkerConfiguration,
    WorkerRun,
    WorkerVersion,
    WorkerVersionGPUUsage,
    WorkerVersionState,
)
from arkindex.project.mixins import ProcessACLMixin, TrainingModelMixin, WorkerACLMixin
from arkindex.training.models import Dataset, Model, ModelVersion
from arkindex.users.models import Role


class StartTrainingSerializer(serializers.ModelSerializer, WorkerACLMixin, TrainingModelMixin):

    name = serializers.CharField(max_length=100, write_only=True)
    corpus_id = serializers.PrimaryKeyRelatedField(
        queryset=Corpus.objects.none(),
        help_text=(
            'ID of the corpus containing the test and train folders. '
            'The user must have a contributor access on the corpus.'
        ),
        style={'base_template': 'input.html'},
        source='corpus',
        write_only=True,
    )
    train_folder_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        help_text='ID of the folder element containing the train dataset.',
        style={'base_template': 'input.html'},
        source='train_folder',
        write_only=True,
    )
    validation_folder_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        help_text='ID of the folder element containing the validation dataset.',
        style={'base_template': 'input.html'},
        source='validation_folder',
        required=False,
        allow_null=True,
        write_only=True,
    )
    test_folder_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        help_text='ID of the folder element containing the test dataset.',
        style={'base_template': 'input.html'},
        source='test_folder',
        required=False,
        allow_null=True,
        write_only=True,
    )
    worker_version_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerVersion.objects.none(),
        help_text=(
            'ID of the worker version to publish a new trained model to. '
            'The user must have an admin access to the worker or repository.'
        ),
        style={'base_template': 'input.html'},
        source='worker_version',
        write_only=True,
    )
    model_id = serializers.PrimaryKeyRelatedField(
        queryset=Model.objects.none(),
        help_text=(
            'ID of the model that will receive the new model version being created by the process. '
            'The user must have a contributor access to the model.'
        ),
        style={'base_template': 'input.html'},
        source='model',
        write_only=True,
    )
    # Use a UUIDField for the model version to ensure it matches the specified model
    model_version_id = serializers.UUIDField(
        help_text=(
            'Optional ID of a model version to be trained again. '
            'The model version must be part of the specified model.'
        ),
        style={'base_template': 'input.html'},
        required=False,
        allow_null=True,
        write_only=True,
    )
    worker_configuration_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerConfiguration.objects.all(),
        help_text='Optional worker configuration to apply on the worker being started',
        style={'base_template': 'input.html'},
        source='worker_configuration',
        required=False,
        allow_null=True,
        write_only=True,
    )
    use_gpu = serializers.BooleanField(default=False, write_only=True)

    # Assign the current user to the process and use Training mode
    mode = serializers.HiddenField(default=ProcessMode.Training)
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())

    class Meta:
        model = Process
        fields = (
            'id',
            'name',
            'corpus_id',
            'train_folder_id',
            'validation_folder_id',
            'test_folder_id',
            'worker_version_id',
            'model_id',
            'model_version_id',
            'worker_configuration_id',
            'use_gpu',
            'mode',
            'creator',
        )
        read_only_fields = ('id',)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'request' not in self.context:
            return

        # This is a hack to make the ACLMixins work on this serializer,
        # because they are designed to be used on APIs and not serializers
        self._user = self.context['request'].user
        corpora = Corpus.objects.writable(self._user)

        # Ensure the user has a contributor access to the corpus and folders
        self.fields['corpus_id'].queryset = corpora
        elements_qs = Element.objects.filter(corpus__in=corpora, type__folder=True)
        self.fields['train_folder_id'].queryset = elements_qs
        self.fields['validation_folder_id'].queryset = elements_qs
        self.fields['test_folder_id'].queryset = elements_qs

        # Restrict the worker version field to all workers that the user can start
        self.fields['worker_version_id'].queryset = (
            WorkerVersion
            .objects
            .filter(worker__in=self.executable_workers)
            # Will be used by WorkerRun.build_summary when starting the process
            .select_related('worker', 'revision')
        )

        # Ensure the user has a contributor access to a potential model
        self.fields['model_id'].queryset = self.editable_models

    def validate_worker_version_id(self, version):
        errors = []
        if version.state != WorkerVersionState.Available:
            errors.append('This worker version is not available.')
        if not version.docker_image_id or not version.docker_image_iid:
            errors.append('This version has no Docker image, which is required for the training.')
        if errors:
            raise ValidationError(errors)
        return version

    def validate(self, data):
        data = super().validate(data)

        # Ensure test, val and train folders are different
        folder_ids = list(
            folder.id
            for folder in (
                data['train_folder'],
                data.get('validation_folder'),
                data.get('test_folder'),
            )
            if folder is not None
        )
        if len(set(folder_ids)) != len(folder_ids):
            raise ValidationError({'__all__': ['Train, validation and test folders must be different.']})

        # Ensure the worker version is compatible with the GPU mode
        if data['worker_version'].gpu_usage == WorkerVersionGPUUsage.Disabled and data["use_gpu"]:
            raise ValidationError({'__all__': ['This worker version does not support GPU mode.']})
        elif data['worker_version'].gpu_usage == WorkerVersionGPUUsage.Required and not data["use_gpu"]:
            raise ValidationError({'__all__': ['This worker version does only support GPU mode.']})

        # Ensure the worker configuration match the given worker
        configuration = data.get('worker_configuration')
        if configuration and configuration.worker_id != data['worker_version'].worker_id:
            raise ValidationError({'__all__': ['The configuration is not compatible with this worker.']})

        # Ensure the model version is part of the defined model and exists
        if data.get('model_version_id'):
            try:
                data['model_version'] = (
                    data['model'].versions
                    # Will be used by WorkerRun.build_summary when starting the process
                    .select_related('model')
                    .get(id=data.pop('model_version_id'))
                )
            except ModelVersion.DoesNotExist:
                raise ValidationError({
                    'model_version_id': ['Object does not exist or is not a version of the specified model']
                })

        return data

    def create(self, validated_data):
        worker_version = validated_data.pop("worker_version")
        worker_configuration = validated_data.pop("worker_configuration", None)
        model_version = validated_data.pop("model_version", None)

        self.instance = super().create(validated_data)

        WorkerRun.objects.create(
            process_id=self.instance.id,
            version=worker_version,
            configuration=worker_configuration,
            parents=[],
            model_version=model_version,
        )

        self.instance.run()
        return self.instance


class ProcessDatasetSerializer(ProcessACLMixin, serializers.ModelSerializer):
    process = serializers.PrimaryKeyRelatedField(
        queryset=(
            Process
            .objects
            # Avoid a stale read when adding a dataset right after creating a process
            .using('default')
            # Required for ACL checks
            .select_related('corpus')
            # Required to check for a process that have already started
            .annotate(has_tasks=Exists(Task.objects.filter(process=OuterRef('pk'))))
        ),
        style={'base_template': 'input.html'},
    )
    dataset = serializers.PrimaryKeyRelatedField(
        queryset=Dataset.objects.none(),
        style={'base_template': 'input.html'},
    )

    class Meta:
        model = ProcessDataset
        fields = ('dataset', 'process', 'id', )
        read_only_fields = ('process', 'id', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return

        request_method = self.context['request'].method
        # Required for the ProcessACLMixin and readable corpora
        self._user = self.context['request'].user

        if request_method == 'DELETE':
            # Allow deleting ProcessDatasets even if the user looses access to the corpus
            self.fields['dataset'].queryset = Dataset.objects.all()
        else:
            self.fields['dataset'].queryset = Dataset.objects.filter(corpus__in=Corpus.objects.readable(self._user))

    def validate_process(self, process):
        access = self.process_access_level(process)
        if not access or not (access >= Role.Admin.value):
            raise PermissionDenied(detail='You do not have admin access to this process.')
        return process

    def validate(self, data):
        process, dataset = data['process'], data['dataset']
        errors = defaultdict(list)

        if process.mode != ProcessMode.Dataset:
            errors['process'].append('Datasets can only be added to or removed from processes of mode "dataset".')

        if process.has_tasks:
            errors['process'].append('Datasets cannot be updated on processes that have already started.')

        if self.context['request'].method == 'POST' and process.datasets.filter(id=dataset.id).exists():
            errors['dataset'].append('This dataset is already selected in this process.')

        if errors:
            raise ValidationError(errors)

        return data
