from collections import OrderedDict

from rest_framework import pagination
from rest_framework.response import Response


class PageNumberPagination(pagination.PageNumberPagination):
    """
    Enhanced PageNumberPagination with extra information to help front-end pagination displays
    """
    page_size_query_param = 'page_size'
    max_page_size = 20

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('count', self.page.paginator.count),
            ('number', self.page.number),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]))

    def get_paginated_response_schema(self, schema):
        schema = super().get_paginated_response_schema(schema)
        schema['properties']['number'] = {
            'type': 'integer',
            'example': 123,
        }
        return schema


class LargePageNumberPagination(PageNumberPagination):
    """
    A numbered pagination with a maximum page size of 500 items.
    The default page size does not change.
    """
    max_page_size = 500


class CustomCursorPagination(pagination.CursorPagination):
    """
    A custom cursor pagination class.
    Ordering may be updated when instantiating the class, but is restricted to fields only on the model and cannot traverse FKs,
    and should be on at least one unique field, as required by DRF:
    https://github.com/encode/django-rest-framework/blob/0ae3323bd237b70d62873fb052cf5617fc446229/rest_framework/pagination.py#L816
    """
    page_size = 20
    page_size_query_param = 'page_size'
    max_page_size = 1000

    def __init__(self, *args, ordering='id', **kwargs):
        self.ordering = ordering
        super().__init__(*args, **kwargs)

    def paginate_queryset(self, queryset, request, *args, **kwargs):
        if request.query_params.get('with_count', '').lower().strip() not in ('', 'false', '0'):
            self.count = queryset.count()
        else:
            self.count = None
        return super().paginate_queryset(queryset, request, *args, **kwargs)

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('count', self.count),
            ('results', data)
        ]))

    def get_schema_operation_parameters(self, view):
        parameters = super().get_schema_operation_parameters(view)
        parameters.append({
            'name': 'with_count',
            'required': False,
            'in': 'query',
            'description': 'Count the total number of elements.',
            'schema': {
                'type': 'boolean',
            }
        })
        return parameters

    def get_paginated_response_schema(self, schema):
        schema = super().get_paginated_response_schema(schema)
        schema['properties']['count'] = {
            'type': 'integer',
            'example': None,
        }
        return schema
