import urllib.parse
from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone

import requests
from django.conf import settings
from django.urls import reverse
from gitlab import Gitlab, GitlabError
from rest_framework.exceptions import AuthenticationFailed, NotAuthenticated


class OAuthProvider(ABC):
    """
    An OAuth authentication provider.
    """

    display_name = ""
    slug = ""

    def __init__(self, credentials=None, url=None):
        if credentials is not None:
            from arkindex.users.models import OAuthCredentials
            assert isinstance(credentials, OAuthCredentials)
        self.credentials = credentials

    @classmethod
    @abstractmethod
    def enabled(cls):
        """
        Boolean stating the provider's availability.
        """

    @abstractmethod
    def get_callback_uri(self):
        """
        Get the OAuth callback URI
        """

    @abstractmethod
    def get_authorize_uri(self):
        """
        Get the OAuth authorization endpoint URI
        """

    @abstractmethod
    def handle_callback(self, request):
        """
        Handle a OAuth callback and save token data. Should raise exceptions if the process fails.
        """

    @abstractmethod
    def refresh_token(self):
        """
        Refresh an expired OAuth token and update OAuthCredentials attributes.
        """

    @abstractmethod
    def disconnect(self):
        """
        Erase token data and logout the user from the service.
        """


class GitLabOAuthProvider(OAuthProvider):

    display_name = 'GitLab'
    slug = 'gitlab'
    default_url = 'https://gitlab.com'
    authorize_endpoint = '/oauth/authorize'
    token_endpoint = '/oauth/token'

    @classmethod
    def enabled(cls):
        return settings.GITLAB_APP_ID and settings.GITLAB_APP_SECRET

    def get_callback_uri(self):
        url = reverse('api:oauth-callback', kwargs={'provider': self.slug})

        if settings.BACKEND_PUBLIC_URL_OAUTH:
            return urllib.parse.urljoin(settings.BACKEND_PUBLIC_URL_OAUTH, url)

        assert settings.PUBLIC_HOSTNAME, 'PUBLIC_HOSTNAME is required to generate callback URIs'
        return urllib.parse.urljoin(settings.PUBLIC_HOSTNAME, url)

    def get_authorize_uri(self):
        if not self.credentials:
            return

        return '{}?{}'.format(
            urllib.parse.urljoin(self.credentials.provider_url, self.authorize_endpoint),
            urllib.parse.urlencode({
                'client_id': settings.GITLAB_APP_ID,
                'redirect_uri': self.get_callback_uri(),
                'scope': 'api',
                'response_type': 'code',
                'state': str(self.credentials.id),
            }),
        )

    def handle_callback(self, request):
        state = request.GET.get('state')
        if not state:
            raise ValueError('No state hash')

        self.credentials = request.user.credentials.get(id=state)

        if not any(param in request.GET for param in ('code', 'error')):
            raise ValueError('Callback called without a valid response')
        if 'error' in request.GET:
            raise ValueError(request.GET.get('error_description', request.GET['error']))

        from arkindex.users.models import OAuthStatus
        assert self.credentials.status != OAuthStatus.Done, 'Cannot overwrite existing credentials'

        self.grant_token(grant_type='authorization_code', code=request.GET.get('code', ''))

    def refresh_token(self):
        try:
            assert self.credentials.refresh_token, 'No refresh token was stored for this OAuth token.'
            self.grant_token(grant_type='refresh_token', refresh_token=self.credentials.refresh_token)
        except Exception as e:
            # Set an error state to allow the user to retry manually
            from arkindex.users.models import OAuthStatus
            self.credentials.status = OAuthStatus.Error
            self.credentials.save()
            raise AuthenticationFailed(
                'The OAuth token could not be refreshed. Please reconnect the OAuth account manually.\n'
                f'Original error: {e}'
            )

    def grant_token(self, **kwargs):
        """
        Use Doorkeeper's OAuth token endpoint to get a token, either from an OAuth authorization code flow
        or when refreshing a token, and update the OAuthCredentials attributes.

        https://github.com/doorkeeper-gem/doorkeeper/wiki/API-endpoint-descriptions-and-examples#post---oauthtoken
        """
        assert self.credentials is not None, 'An OAuthCredentials instance is required to use the OAuth token endpoint'

        payload = {
            'client_id': settings.GITLAB_APP_ID,
            'client_secret': settings.GITLAB_APP_SECRET,
            'redirect_uri': self.get_callback_uri(),
        }
        payload.update(kwargs)
        response = requests.post(
            urllib.parse.urljoin(self.credentials.provider_url, self.token_endpoint),
            payload,
        )
        response.raise_for_status()
        data = response.json()

        assert 'access_token' in data, 'GitLab returned no OAuth token'
        assert data.get('token_type', '').lower() == 'bearer', 'Unsupported OAuth token type'

        self.credentials.token = data['access_token']
        self.credentials.refresh_token = data.get('refresh_token')
        # Get absolute date of expiration from the created_at UNIX timestamp and the seconds until expiration
        self.credentials.expiry = datetime.fromtimestamp(data['created_at'], tz=timezone.utc) + timedelta(seconds=int(data['expires_in']))

        gl = Gitlab(self.credentials.provider_url, oauth_token=self.credentials.token)
        gl.auth()
        self.credentials.account_name = gl.user.username

        from arkindex.users.models import OAuthStatus
        self.credentials.status = OAuthStatus.Done
        self.credentials.save()

    def disconnect(self):
        if not self.credentials:
            raise NotAuthenticated

        if not self.credentials.token or not self.credentials.repos.exists():
            return

        # Remove all webhooks
        try:
            gl = Gitlab(self.credentials.provider_url, oauth_token=self.credentials.token)
            for repo in self.credentials.repos.all():
                project = gl.projects.get(urllib.parse.urlsplit(repo.url).path.strip('/'))
                hook_url = urllib.parse.urljoin(
                    settings.PUBLIC_HOSTNAME,
                    reverse('api:import-hook', kwargs={'pk': repo.id})
                )
                # Try to find the webhook
                hook = next((h for h in project.hooks.list() if h.url == hook_url), None)
                if hook:
                    hook.delete()
        except GitlabError:
            pass


oauth_providers = [
    GitLabOAuthProvider,
]


def get_provider(name):
    return next(filter(lambda p: p.slug == name, oauth_providers), None)
